import numpy as np
from skimage.draw import random_shapes
from skimage.segmentation import relabel_sequential

from fast_overlap import overlap, overlap_parallel

im_shape = (1024, 1024)
min_shapes = 5
im1 = random_shapes(im_shape, 20, min_shapes=min_shapes, random_seed=0)[0]
im1 = relabel_sequential(im1.sum(axis=-1))[0]
im2 = random_shapes(im_shape, 20, min_shapes=min_shapes, random_seed=1995)[0]
im2 = relabel_sequential(im2.sum(axis=-1))[0]
shape = (int(np.max(im1) + 1), int(np.max(im2) + 1))
out = overlap_parallel(im1.astype(np.int32), im2.astype(np.int32), shape)
print(out.sum())
# out_serial = overlap(im1.astype(np.int32), im2.astype(np.int32), shape)
out_serial = overlap(im1.astype(np.uint16), im2.astype(np.uint16), shape)

# from IPython import embed
# embed(colors="Linux")
assert np.all(out == out_serial)



# distutils: language=c++

import numpy as np

cimport cython
cimport numpy as np

__all__ = [
    "overlap",
    "overlap_parallel",
]
cimport numpy as np


cdef extern from "parallel_overlap.cpp":
    cdef void overlap_parallel_cpp(int *, int *, Py_ssize_t[2], int *, Py_ssize_t) nogil

@cython.wraparound(False)
@cython.nonecheck(False)
cpdef overlap_parallel(int [:,::1] prev, int[:,::1] curr, shape=None):
    """
    Calculate the pairwise overlap the labels for two arrays in
    parallel using openmp. This may not work on non-linux OSes.

    Currently limited to only accept `int` type arrays.

    Parameters
    ----------
    prev, curr : 2D array-like of int
        curr will have at least as many unique labels as prev
    shape : tuple of int, optional
        The shape of the output array. This should reflect the maximum
        value of labels.

    Returns
    -------
    arr : (N, M) array of int
        N is the number of unique labels in prev and M the number of unique in curr.
        The ijth entry in the array gives the number of pixels for which label *i* in prev
        overlaps with *j* in curr.
    """
    prev = np.ascontiguousarray(prev)
    curr = np.ascontiguousarray(curr)

    if shape is None:
        shape = (prev.max(), curr.max())

    cdef np.ndarray[int, ndim=2, mode="c"] output = np.zeros(shape, dtype=np.dtype("i"))
    cdef Py_ssize_t ncols = shape[1]

    with nogil:
        overlap_parallel_cpp(&prev[0,0], &curr[0,0], prev.shape, &output[0,0], ncols)
    return output



from libc cimport stdint

ctypedef fused ints:
    stdint.uint8_t
    stdint.uint16_t
    stdint.uint32_t
    stdint.uint64_t
    stdint.int8_t
    stdint.int16_t
    stdint.int32_t
    stdint.int64_t

# @cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cpdef overlap(ints[:, :] prev, ints[:,:] curr, shape=None):
    """
    Calculate the pairwise overlap the labels for two arrays.

    Parameters
    ----------
    prev, curr : 2D array-like of int
        curr will have at least as many unique labels as prev
    shape : tuple of int, optional
        The shape of the output array. This should reflect the maximum
        value of labels.

    Returns
    -------
    arr : (N, M) array of int
        N is the number of unique labels in prev and M the number of unique in curr.
        The ijth entry in the array gives the number of pixels for which label *i* in prev
        overlaps with *j* in curr.
    """
    prev = np.ascontiguousarray(prev)
    curr = np.ascontiguousarray(curr)

    if shape is None:
        shape = (np.max(prev)+1, np.max(curr)+1)

    cdef int [:, :] arr
    arr = np.zeros(shape, dtype=np.dtype("i"))
    cdef size_t I, J, i, j
    for i in range(prev.shape[0]):
        for j in range(prev.shape[1]):
            p = prev[i,j]
            c = curr[i,j]
            if p and c:
                arr[p, c] += 1
    return np.asarray(arr)



from timeit import timeit

setup = """
from fast_overlap import overlap_parallel, overlap

from skimage.draw import random_shapes
from skimage.segmentation import relabel_sequential
import numpy as np

im_shape=(1024,1024)
min_shapes = 5
im1 =random_shapes(im_shape, 20, min_shapes=min_shapes, random_seed=0)[0]
im1 = relabel_sequential(im1.sum(axis=-1))[0]
im2 =random_shapes(im_shape, 20, min_shapes=min_shapes, random_seed=1995)[0]
im2 = relabel_sequential(im2.sum(axis=-1))[0]
shape = (int(np.max(im1)+1), int(np.max(im2)+1))

from numba import int32, jit

@jit((int32[:,:], int32[:,:], int32, int32), nopython=True)
def overlap_numba(prev, curr, shape1, shape2):
    arr = np.zeros((shape1, shape2), dtype=np.dtype("i"))
    for i in range(prev.shape[0]):
        for j in range(prev.shape[1]):
            arr[prev[i,j],curr[i,j]] += 1
    return arr
"""

# import matplotlib.pyplot as plt
# fig, axs = plt.subplots(1,2)
# axs[0].imshow(im1)
# axs[1].imshow(im2)
# plt.show()
times = {}
times["serial_cpp"] = timeit(
    "overlap(im1.astype(np.int32), im2.astype(np.int32), shape)", setup=setup, number=10
)
times["parallel_cpp"] = timeit(
    "overlap_parallel(im1.astype(np.int32), im2.astype(np.int32), shape)",
    setup=setup,
    number=10,
)
times["serial_numba"] = timeit(
    "overlap_numba(im1.astype(np.int32), im2.astype(np.int32), *shape)",
    setup=setup,
    number=10,
)
print(times)



# noqa: E402
import os
import sys

if os.environ.get("CI", False):
    # remove the local version from sys.path so that we use the version
    # we installed from the wheel
    sys.path = sys.path[1:]

from pathlib import Path  # noqa: E402

import numpy as np  # noqa: E402
import pytest  # noqa: E402

import fast_overlap  # noqa: E402

ims = np.load(str(Path(__file__).parent / "test-ims.npy")).astype(np.int64)
expected = np.load(str(Path(__file__).parent / "expected-overlap.npy"))
shape = (int(np.max(ims[0]) + 1), int(np.max(ims[1]) + 1))


# test a few different types but not all
@pytest.mark.parametrize("type", [np.uint16, np.uint64, np.int32, np.int64])
def test_overlap(type):
    out = fast_overlap.overlap(ims[0].astype(type), ims[1].astype(type), shape)
    assert np.all(out == expected)


def test_overlap_no_shape():
    out = fast_overlap.overlap(ims[0], ims[1])
    assert np.all(out == expected)


def test_parallel_overlap():
    out = fast_overlap.overlap_parallel(
        ims[0].astype(np.int32), ims[1].astype(np.int32), shape
    )
    assert np.all(out == expected)



import os

import numpy
from Cython.Build import cythonize
from setuptools import Extension, setup

# os.environ["CC"] = "C:/Program Files/LLCM/bin/clang-cl.exe"
# os.environ["CXX"] = "C:/Program Files/LLCM/bin/clang-cl.exe"
if os.name == "nt":
    compile_args = ["/fopenmp", "/Ox"]
    extra_link_args = []
else:
    compile_args = [
        "-fopenmp",
        "-O3",
    ]
    extra_link_args = ["-fopenmp"]
# if os.name == "darwin":
#     compile_args.append("-lomp")
setup(
    name="fast_overlap",
    use_scm_version={"write_to": "fast_overlap/_version.py"},
    ext_modules=cythonize(
        [
            Extension(
                "fast_overlap._engine",
                ["fast_overlap/fast_overlap.pyx"],
                include_dirs=[numpy.get_include()],
                extra_compile_args=compile_args,
                extra_link_args=extra_link_args,
            )
        ],
        language_level="3",
    ),
)



try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"
from ._engine import overlap, overlap_parallel

__all__ = ["__version__", "overlap", "overlap_parallel"]
